#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2014 Gary Wright wrightsolutions.co.uk/contact
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Tested against fabric 1.4.3
# http://ftp.uk.debian.org/debian/pool/main/f/fabric/fabric_1.4.3-1.debian.tar.gz

from __future__ import with_statement
from fabric.api import abort, cd, env, get, put, run, settings, sudo
from fabric.contrib.console import confirm
from fabric.contrib.files import exists,append
from os import path
import time
#from collections import namedtuple,OrderedDict
from . import cmdbin
from . import files_directories as files_dirs
from . import package_deb as pdeb
from . import sign_repo

""" res.succeeded is preferred for newer fabric, but not res.failed
is used where backward compatibility is preferable. """

PORT_WSGI = '1480'
""" Above stem is 14 to represent Django 14 but change to 1680 when Django 16
Originally PORT_WSGI was 8080 but we are making an effort to avoid clashing with
an existing typical web server port assignment so benchmarking less likely to
interfere with exising setup
"""
GROUP_WRITE_EXT_LIST_WEB = ['bz2','gz','log','lz','tar','zip']
GROUP_WRITE_EXT_LIST_PYWEB = GROUP_WRITE_EXT_LIST_WEB[:]
GROUP_WRITE_EXT_LIST_PYWEB.extend(['db3','sqlite','sqlite3'])
# DIRSTEM_DATA_SHORT = '~/'
DIRSTEM_DATA_SHORT = '/var/local'
GROUP_WEBSERVER = 'www-data'

def group_writable_web(dirpath,group_write_ext_list=None):
    if group_write_ext_list is None:
        group_write_ext_list = GROUP_WRITE_EXT_LIST_WEB
    return files_dirs.group_writable(dirpath,'/var/www/pub','/var/l',group_write_ext_list)

def group_writable_web_recursive(dirpath,ext_list=None):
    if ext_list is None:
        ext_list = GROUP_WRITE_EXT_LIST_WEB
    return files_dirs.group_writable_recursive(dirpath,path_startswith1='/var/www/pub',
                          path_startswith2=None,group_write_ext_list=ext_list)

def group_writable_pyweb(dirpath,group_write_ext_list=None):
    if group_write_ext_list is None:
        group_write_ext_list = GROUP_WRITE_EXT_LIST_PYWEB
    return files_dirs.group_writable(dirpath,'/var/www/pub','/var/l',group_write_ext_list)

def group_writable_pyweb_recursive(dirpath,ext_list=None):
    if ext_list is None:
        ext_list = GROUP_WRITE_EXT_LIST_PYWEB
    return files_dirs.group_writable_recursive(dirpath,path_startswith1='/var/www/pub',
                          path_startswith2=None,group_write_ext_list=ext_list)

def chgrp_or_sudo_web(dirpath,newgroup=None):
    if newgroup and len(newgroup) > 1:
        grp = newgroup
    else:
        grp = GROUP_WEBSERVER
    return files_dirs.chgrp_or_sudo(dirpath,grp)


def apache2threads(apache2user='www-data',print_flag=False):
    threads = cmdbin.process_user(apache2user)
    if print_flag is not False:
        for tline in threads:
            print tline
    return threads


def apache2mpm_lines(print_flag=False):
    mpm_lines = []
    with settings(warn_only=True):
        mres = cmdbin.run_or_sudo_warningless(cmdbin.CMD_APACHE_MPM_GREPPED)
        mpm_lines = cmdbin.run_res_lines_stripped(mres)
    if print_flag is not False:
        for mline in mpm_lines:
            print mline
    return mpm_lines


def apache2mpm():
    mpm = None
    mpm_line = ''
    with settings(warn_only=True):
        mres = cmdbin.run_or_sudo_warningless(cmdbin.CMD_APACHE_V)
        if mres.succeeded:
            mpm = 'installed'
            mpm_line = cmdbin.run_res_lines_stripped_searching1(mres,'MPM:')
    if len(mpm_line) > 7:
        # chr(58) is colon (:)
        try:
            line_lower_array = mpm_line.lower().split(chr(58))
            mpm = line_lower_array[-1].strip()
        except:
            mpm = 'unknown'
    return mpm


def manage_py_or_sudo(manage_command='syncdb',dirproject=None):
    managed_it_flag = False
    manage_dir = None
    if dirproject:
        manage_dir = dirproject
    else:
        try:
            if files_dirs.slash_count(dirpath_django_project) > 0:
                manage_dir = dirpath_django_project
        except Exception:
            manage_dir = None

    if manage_dir and files_dirs.slash_count(manage_dir) > 0:
        with cd(manage_dir):
            manage_line = "./manage.py {0}".format(manage_command)
            if cmdbin.run_or_sudo(manage_line):
                managed_it_flag = True
    return managed_it_flag


def sync_load_index(dirpath_project,dirpath_log,dirpath_djapian,print_flag=False):

    sli_elapsed = 0
    t1 = time.time()
    sync_res = manage_py_or_sudo('syncdb',dirpath_project)

    load_res = manage_py_or_sudo('loaddata local_hyphen_source_def',dirpath_project)

    perm_res = files_dirs.group_writable(dirpath_log,'/var/www/pub','/var/l')

    index_res = manage_py_or_sudo('index --rebuild',dirpath_project)
    """ Multiple runs of --rebuild might produce spurious errors as it is
    only supposed to be used the first time the index is built. 
    This is why there is a directory removal earlier in the script to try
    and avoid such issue where multiple runs / re-run of this benchmark.
    """

    group_write_res = group_writable_pyweb_recursive(dirpath_djapian)
    t2 = time.time()
    """ Above gives appropriate access on Djapian subdirectories which
    helps to avoid DatabaseLockError: Unable to get write lock on ...
    """
    if not group_write_res:
        files_dirs.abort_with_message('Aborting during Djapian index permission setting.')

    sli_elapsed = t2-t1

    if print_flag:
        time_line = "sync, load, index timing: {0:.3f} seconds".format(sli_elapsed)
        print cmdbin.lavg_prefixed(time_line)

    return sli_elapsed


def dump_repeat_hash(dirpath_bin,dirpath_project):
    manage_line = "{0}/manage_dumpdata_million_repeat_hash.sh {1}".format(
        dirpath_bin,dirpath_project)
    drh_elapsed = 0
    drh1 = time.time()
    drh_res = cmdbin.run_or_sudo(manage_line)
    if drh_res:
        drh2 = time.time()
        drh_elapsed = drh2-drh1
    int_drh_elapsed = int(drh_elapsed)
    return int_drh_elapsed


def poor_cycle_setup14(target='/rc/py',concurrent=2,totaltime=180,
                       data='initial',visible='internal',
                       app_name='local2dj14',dir_fab='poor_cycle14.fab',port_wsgi=None):
    path3 = none3 = (None,None,None)

    (pname, pretty_name_id, pretty_name_word) = cmdbin.pretty_name_tuple()

    print pretty_name_id
    print pretty_name_word

    """ The next code block is just a helper to allow shortened rc/py
    as argument. """
    if not target:
        pass
    elif target.startswith('prettify/'):
        pass
    elif target == 'rc/py/':
        target = 'prettify/local-database/rc/py/'
    elif target == 'rc/py':
        target = 'prettify/local-database/rc/py'
    elif target.endswith('/rc/py/'):
        target = 'prettify/local-database/rc/py/'
    elif target.endswith('/rc/py'):
        target = 'prettify/local-database/rc/py'

    dirstem_data = files_dirs.tilde_expand(DIRSTEM_DATA_SHORT)
    dirpath_data = "{0}/django14data".format(dirstem_data)
    mkdir_res = files_dirs.mkdir_existing("{0}/django14data".format(DIRSTEM_DATA_SHORT),
                               2771,False)
    """ mkdir_res = run("mkdir -m 2771 {0}".format(
            "{0}/django14data".format(dirstem_data)))
    """
    if mkdir_res:
        files_dirs.dirpath_dlp(dirpath_data)
        #print_success_and_failed(mkdir_res,dirpath_data)
    else:
        pass
        """ abort_with_cleanup(dirpath_fab,'Aborting during directory initialisation.') """
        # abort('Aborted during initialisation of directory django14data.')

    chown_line = "{0} root:www-data {1}".format(files_dirs.CMD_CHOWN,dirpath_data)
    chown_res = cmdbin.run_or_sudo(chown_line)
    cmdbin.print_success_and_failed(chown_res,dirpath_data)
    if chown_res.failed:
        abort('Aborted during initialisation of directory django14data ownership.')


    dirpath_djapian_app = "{0}/djapian/{1}".format(dirstem_data,app_name)
    rmdir_res = files_dirs.rmdir_existing(dirpath_djapian_app,True)
    if not rmdir_res:
        pass
        """ abort_with_cleanup(dirpath_fab,'Aborting during directory initialisation.') """
        #abort('Aborting during remove of djapian indexes for app.')

    mkdir_res = files_dirs.mkdir_existing("{0}/djapian".format(DIRSTEM_DATA_SHORT),2771,True,True,True)
    if not mkdir_res:
        """ abort_with_cleanup(dirpath_fab,'Aborting during directory initialisation.') """
        abort('Aborting during initialisation of directory djapian.')

    """ Read and understand function mkdir_existing and safety features to see why generally
    it is preferred to send any tilde or short form of directory to that function.
    That knowledge helps explain why dirpath_djapian is defined here rather than being
    used in the call to mkdir_existing() above.
    """
    dirpath_djapian = "{0}/djapian".format(dirstem_data)

    """ Now above we gave group write via 2771, but here we need to set the group to which
    that group write privilege will belong. """
    chgrp_or_sudo_web(dirpath_djapian)

    if env.user == 'root':
        files_dirs.set_dirpath_fab_from_short("/var/www/public_html/{0}".format(dir_fab))
    else:
        files_dirs.set_dirpath_fab_from_short("~/public_html/{0}".format(dir_fab))

    externally_visible_flag = False
    if 'both' in visible:
        externally_visible_flag = True

    repo_url_django_name = 'local2django14bench'
    repo_url_django = "https://bitbucket.org/wrightsolutions/{0}".format(
        repo_url_django_name)
    repo_url_localbin = "https://bitbucket.org/wrightsolutions/local-bin"
    repo_url_localwebserver = "https://bitbucket.org/wrightsolutions/local-webserver"

    dirpath_fab = files_dirs.get_dirpath_fab()
    """ Make temporary directory in user home to hold cloned repo """
    mkdir_res = files_dirs.mkdir_existing(dirpath_fab,2755,True,True,True)
    if mkdir_res:
        print "Created directory {0}".format(dirpath_fab)
    else:
        abort('Aborted during initialisation of .fab directory.')

    print "Cloning into directory {0}".format(dirpath_fab)
    with cd(dirpath_fab):
        sign_repo.repo_clone(repo_url_django,'fd96e9c')
        sign_repo.repo_clone(repo_url_localbin,'dcbd2b2057cdce99fb19fee5b711e7f24ed52923')
        sign_repo.repo_clone(repo_url_localwebserver,'8647d7b821646a52c526e6096ccdf12588ef0280')

    requires_django_res = pdeb.requires_binary_package('django-admin',
                                                       'python-django',assumeyes=True)
    if requires_django_res:
        requires_django_which = requires_django_res

    requires_a2enmod_res = pdeb.requires_binary_package('a2enmod',
                                                        'apache2.2-common',assumeyes=True)
    """ Although we do not use a2enmod directly (just yet) we do need to ensure, ahead of time,
    that common /etc/apache2 directories will exist. Pushing a file to mods-available should
    then not create an issue """

    dirstem_django = dirpath_fab
    dirpath_django_project = "{0}/{1}".format(dirstem_django,
                                              repo_url_django_name)
    dirpath_django_app = "{0}/{1}".format(dirpath_django_project,app_name)

    mpm = apache2mpm()
    print "Current apache2mpm is {0}".format(mpm)
    if mpm == 'worker':
        worker_better = 'which might result in better benchmark results'
        if not confirm("Worker MPM setup in place {0}. Continue anyway?".format(worker_better)):
            return none3

    """ Ensure apache2 binary is available as otherwise our reload / restart will fail later """
    requires_apache2_res = pdeb.requires_binary_package('apache2',
                                                        'apache2-mpm-prefork',assumeyes=True)

    requires_wsgi_res = pdeb.requires_dpkg_package('libapache2-mod-wsgi',True)
    """ Above requires should come before we send any templated wsgi related conf to the server """

    path_line = """
<IfModule mod_wsgi.c>
WSGIPythonPath {0}:{1}
</IfModule>""".format(dirstem_django,dirpath_django_project)
    # .format(dirstem_django,repo_url_django_name)
    if env.user == 'root':
        append('/etc/apache2/mods-available/wsgi.conf',path_line)
    else:
        append('/etc/apache2/mods-available/wsgi.conf',path_line,use_sudo=True)

    from jinja2 import Environment

    local_settings_tpl = """
# Django local_settings for {{ project }} project.
DEBUG = False
TEMPLATE_DEBUG = DEBUG

ADMINS = (
    # ('Your Name', 'your_email@example.com'),
)

MANAGERS = ADMINS

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3', # Add 'postgresql_psycopg2', 'mysql', 'sqlite3' or 'oracle'.
        'NAME': '{{ dbpath }}/{{ dbname }}.sqlite', # Or path to database file if using sqlite3.
        'USER': '',                      # Not used with sqlite3.
        'PASSWORD': '',                  # Not used with sqlite3.
        'HOST': '',                      # Set to empty string for localhost. Not used with sqlite3.
        'PORT': '',                      # Set to empty string for default. Not used with sqlite3.
    }
}

LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse'
        }
    },
    'formatters': {
        'verbose': {
            'format': '%(levelname)s %(asctime)s %(module)s %(process)d %(thread)d %(message)s'
        },
        'simple': {
            'format': '%(levelname)s %(message)s'
        },
    },
    'handlers': {
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler'
        },
        'log_file': {
            'level': 'DEBUG',
            'class': 'logging.handlers.RotatingFileHandler',
            'formatter': 'verbose',
            'filename': '{{ djangolog }}',
            'maxBytes': 4194304, # 4 * 1048576
            'backupCount': 5,
        },
    },
    'loggers': {
        '': {
            'handlers': ['log_file'],
            'level': 'DEBUG',
            'propagate': True
        },
        'django.request': {
            'handlers': ['log_file'],
            'level': 'WARNING',
            'propagate': False,
        },
    }
}

DJAPIAN_DATABASE_PATH = '{{ dirdjapian }}'
"""

    dirpath_django_log = "{0}/local2.log".format(dirpath_data)

    path3 = (dirpath_django_project,dirpath_django_log,dirpath_djapian)

    template_dict = {
        'dbpath': dirstem_data,
        'dbname': repo_url_django_name,
        'djangolog': dirpath_django_log,
        'dirdjapian': dirpath_djapian,
        'project': repo_url_django_name,
        }
    local_settings = Environment().from_string(local_settings_tpl).render(template_dict)

    files_dirs.put_using_named_temp("{0}/local_settings.py".format(dirpath_django_project),
                         local_settings,'local_settings construction',False)

    wsgi_py = """
# WSGI config for {0} project.
from os import environ

environ.setdefault("DJANGO_SETTINGS_MODULE", "{1}")

# This application object is used by any WSGI server configured to use this
# file. This includes Django's development server, if the WSGI_APPLICATION
# setting points here.
                
from django.core.wsgi import get_wsgi_application
application = get_wsgi_application()
""".format(repo_url_django_name,'settings')

    files_dirs.put_using_named_temp("{0}/wsgi.py".format(dirpath_django_app),
                         wsgi_py,'wsgi.py construction',False)

    requires_djapian_res = pdeb.requires_dpkg_package('python-django-djapian',
                                                      True,'universe')
    if not requires_djapian_res:
        files_dirs.abort_with_message('Aborting during djapian availability check.')

    requires_yaml_res = pdeb.requires_dpkg_package('python-yaml',True)
    if not requires_yaml_res:
        files_dirs.abort_with_message('Aborting during yaml availability check.')

    requires_pygments_res = pdeb.requires_dpkg_package('python-pygments',True)
    if not requires_pygments_res:
        files_dirs.abort_with_message('Aborting during pygments availability check.')

    apache2conf_tpl = """
<VirtualHost *:{{ port_host }}>

DocumentRoot {{ docroot_slashless }}

<Directory />
   Options -Indexes
   AllowOverride None
   Order allow,deny
   allow from all
</Directory>

<Directory "{{ docroot_slashless }}/">
   Options -Indexes
   AllowOverride None
   Order allow,deny
   allow from all
</Directory>

   WSGIDaemonProcess {{ app }} user=www-data group=www-data processes=1 threads=5
   WSGIScriptAlias / {{ docroot_slashless }}/{{ project }}/{{ app }}/wsgi.py
   #WSGIPythonPath /opt/local/webdata/local2bottle/wsgi/# wsgi.conf

   Alias /static {{ docroot_slashless }}/{{ project }}/{{ app }}/static
   #Alias /static {{ docroot_slashless }}/local2django14/local2dj14/static
</VirtualHost>
"""

    if port_wsgi is None:
        port_wsgi = PORT_WSGI
    else:
        try:
            port_wsgi = int(port_wsgi)
            if port_wsgi < 80:
                port_wsgi = PORT_WSGI
        except:
            port_wsgi = PORT_WSGI

    template_dict = {
        'port_host': port_wsgi,
        'docroot_slashless': dirpath_fab,
        'project': repo_url_django_name,
        'app': app_name,
        }

    apache2conf = Environment().from_string(apache2conf_tpl).render(template_dict)
    # docroot_slashless=dirstem_data + '/django14')

    if port_wsgi in [80,8080]:
        continue_add_port = 'Continue and add port'
        if not confirm("Apache will listen additionally on port {0}. {1}?".format(
                port_wsgi,continue_add_port)):
            return none3

    apache2conf_res = None
    import tempfile
    named_temp = tempfile.NamedTemporaryFile()
    try:
        named_temp.write(apache2conf)
        named_temp.flush()
        apache2conf_res = put(named_temp,"/etc/apache2/sites-enabled/{0}".format(app_name),
                              use_sudo=True)
    finally:
        named_temp.close()

    if externally_visible_flag:
        listen_line = "Listen {0}".format(port_wsgi)
    else:
        listen_line = "Listen 127.0.0.1:{0}".format(port_wsgi)

    port_lines = """
<IfModule mod_wsgi.c>
# NameVirtualHost *:{0}
{1}
</IfModule>
""".format(port_wsgi,listen_line)

    named_temp = tempfile.NamedTemporaryFile()
    try:
        named_temp.write(port_lines)
        named_temp.flush()
        apache2port_res = put(named_temp,'/etc/apache2/ports.conf.append',use_sudo=True)
    finally:
        named_temp.close()

    if apache2port_res:
        if apache2conf_res.failed:
            files_dirs.abort_with_message('Aborting during apache2 port put.')
    else:
            files_dirs.abort_with_message('Aborting during apache2 port.')

    append_bin = "{0}/local-bin/append_not_already_dot_append.sh".format(dirpath_fab)
    append_run = "{0} /etc/apache2/ports.conf 'Listen.*{1}'".format(append_bin,port_wsgi)
    append_res = run(append_run)
    if append_res.failed:
        files_dirs.abort_with_message('Aborting during apache2 port addition.')

    if apache2conf_res:
        if apache2conf_res.failed:
            files_dirs.abort_with_message('Aborting during apache2 configuration put.')
    else:
            files_dirs.abort_with_message('Aborting during apache2 configuration.')

    """ apache2re_res = sudo("/etc/init.d/apache2 reload",group='adm') """
    apache2re_res = sudo("/etc/init.d/apache2 reload")
    if apache2re_res.failed:
        files_dirs.abort_with_message('Aborting during apache2 reload.')

    if 'root' in env.user:
        pass
    else:
        apache2enmod_res = sudo("/usr/sbin/a2enmod userdir")
        if apache2enmod_res.failed:
            files_dirs.abort_with_message('Aborting during apache2 Enable Module userdir.')

    """ apache2re_res = sudo("/etc/init.d/apache2 reload") """
    restart_bin = "{0}/{1}/apache2/apache2stop5startconditional.sh".format(
        dirpath_fab,'local-webserver')
    apache2re_res = sudo(restart_bin)
    if apache2re_res.failed:
        files_dirs.abort_with_message('Aborting during apache2 stop 5 start.')

    with settings(warn_only=True):
        apache2port_res = sudo("{0} :{1}".format(cmdbin.CMD_LSOF_INTERNET,port_wsgi))
        if apache2port_res.failed:
            files_dirs.abort_with_message('Aborting during apache2 port test.')

    with settings(warn_only=True):
        url_cdn = 'http://c5871e0b1f1c144e2f0e-7fd212704b537f6e17de67217009ee51.r73.cf3.rackcdn.com'
        url_filename = 'constructionMStankieA14znak20111228164215derivativePublicDomainPx125.png'
        with cd(dirpath_fab):
            run(cmdbin.CMD_WGET + "{0}/{1}".format(url_cdn,url_filename))
            # if png_res.failed:

    index_html = """
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
        "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" lang="en" xml:lang="en">
<head>
<meta http-equiv="Content-type" content="text/html;charset=UTF-8" />
<meta name="robots" content="noarchive,follow" />
<title>Under Construction</title>
</head>
<body>
<p><img src="{0}" alt="Under Construction Logo" /> coming soon...</p>
<p>
    <a href="http://validator.w3.org/check?uri=referer"><img
        src="http://www.w3.org/Icons/valid-xhtml10-blue"
        alt="Valid XHTML 1.0 Strict" height="31" width="88" /></a>
  </p>
</body>
</html>
""".format(url_filename)

    index_path = "{0}/index.html".format(dirpath_fab)
    with cd(dirpath_fab):
        files_dirs.put_using_named_temp(index_path,index_html,'index.html construction',False)

    return path3


def poor_cycle_load14(target='/rc/py',concurrent=2,totaltime=10,dir_fab='poor_cycle14.fab',
                      dirpath_django_project=None,dirpath_django_log=None,
                      dirpath_djapian=None,port_wsgi=None):
    """ If you want fabric command line interaction with an argument
    then expect it to be a string. May need to int() what you get """
    ab_concurrent = int(concurrent)
    ab_totaltime = int(totaltime)
    tuple5 = zero5 = (0,0,0,0,0)
    if ab_totaltime < 1:
        return zero5

    requires_ab_res = None
    requires_ab_which = cmdbin.CMD_FREE
    if int(totaltime) > 0:
        requires_ab_res = pdeb.requires_binary_package('ab','apache2-utils',assumeyes=True)
        if requires_ab_res:
            requires_ab_which = requires_ab_res

    import time
    c1 = time.time()
    time_line_dict = {}
    """ seconds is not necessarily unique and so seconds is the VALUE not the key """
    seconds1 = sync_load_index(dirpath_django_project,dirpath_django_log,dirpath_djapian,True)
    time_line1 = "(1) sync, load, index timing: {0:.3f} seconds".format(seconds1)
    time_line_dict[time_line1]=seconds1

    seconds2 = sync_load_index(dirpath_django_project,dirpath_django_log,dirpath_djapian,True)
    time_line2 = "(2) sync, load, index timing: {0:.3f} seconds".format(seconds2)
    time_line_dict[time_line2]=seconds2

    seconds3 = sync_load_index(dirpath_django_project,dirpath_django_log,dirpath_djapian,True)
    time_line3 = "(3) sync, load, index timing: {0:.3f} seconds".format(seconds3)
    time_line_dict[time_line3]=seconds3

    seconds4 = sync_load_index(dirpath_django_project,dirpath_django_log,dirpath_djapian,True)
    time_line4 = "(4) sync, load, index timing: {0:.3f} seconds".format(seconds4)
    time_line_dict[time_line4]=seconds4
    #print cmdbin.lavg1prefixed(time_line4)

    print 'Sync, load, index output text repeated but in ascending order (of seconds) next'
    """ time_key next will be the time_line (text) as documented earlier """
    for time_key,time_value in sorted(time_line_dict.items(), key=lambda t: t[1]):
        print time_key
    print 'Sync, load, index output text (ordered) that shows (1), (2), (3), (4) might indicate degradation?'

    if ab_totaltime > 5:
        #if env.user == 'root':
        #    files_dirs.set_dirpath_fab_from_short("/var/www/public_html/{0}".format(dir_fab))
        #else:
        #    files_dirs.set_dirpath_fab_from_short("~/public_html/{0}".format(dir_fab))

        dirpath_fab = files_dirs.get_dirpath_fab()
        if dirpath_fab is None:
            print 'directory=None ? Did benchmark setup (poor_cycle_setup14 or poor_cycle_setup16) complete?'
            return zero5

        dump_elapsed = dump_repeat_hash("{0}/local-bin/".format(dirpath_fab),dirpath_django_project)
        if dump_elapsed < 1:
            print 'Benchmark may be suspect - dump_repeat_hash() completed in less < 1 sec'
    else:
        print "Benchmark step dump_repeat_hash() skipped for requested totaltime={0}".format(ab_totaltime)

    # By making final paramater True instead of False below, you can delegate the print instead
    sync_load_index(dirpath_django_project,dirpath_django_log,dirpath_djapian,False)
    time_line5 = "(1) sync, load, index timing: {0:.3f} seconds".format(seconds1)
    print cmdbin.lavg_prefixed(time_line5)

    c2 = time.time()

    cycle_elapsed = c2-c1
    time_complement = 0
    try:
        int_cycle_elapsed = int(cycle_elapsed)
        print "Balance calculation (seconds) will involve ({0} - {1})".format(
            ab_totaltime,int_cycle_elapsed)
        time_complement = ab_totaltime - int_cycle_elapsed
        if time_complement < 5:
            time_complement = 5
            print "Balance calculation (seconds) has prompted an adjustment time_complement=".format(
                time_complement)
    except Exception,e:
        print "Balance calculation (seconds) not completed! {0}".format(e)
        time_complement = 0

    if port_wsgi is None:
        port_wsgi = PORT_WSGI
    else:
        try:
            port_wsgi = int(port_wsgi)
            if port_wsgi < 80:
                port_wsgi = PORT_WSGI
        except:
            port_wsgi = PORT_WSGI

    url_counts = "http://127.0.0.1:{0}/local2/counts/".format(port_wsgi)
    url_benchmark = "http://127.0.0.1:{0}/local2/{1}".format(port_wsgi,target)

    if requires_ab_res and ab_totaltime > 1:
        """ To get things primed, do a single fetch first then short sleep. """
        run("{0} {1}".format(cmdbin.CMD_WGET1,url_counts))
        if time_complement > 0:
            time.sleep(2)
            print "Using time_complement={0} to complete next stage /usr/bin/ab".format(
                time_complement)
            """ Benchmark the url """
            """ run("/usr/bin/ab -kc {0:d} -t {1:d}".format(ab_concurrent,time_complement)) """
            run("{0} -kc {1:d} -t {2:d} {3}/".format(requires_ab_which,ab_concurrent,
                                                     time_complement,url_benchmark))
    elif ab_totaltime == 1:
        """ In this context we use timeout 1 for our single fetch """
        print 'Fetch: In this context we use timeout 1 for our single fetch'
        run("{0} {1}".format(cmdbin.CMD_WGET1,url_benchmark))
    else:
        """ No --timeout flag given to wget here """
        print 'Fetch: In this context we use no timeout for our single fetch'
        run("{0} {1}".format(cmdbin.CMD_WGET,url_benchmark))
    load3 = cmdbin.load3tuple()
    if load3[0] is not None:
        tuple5 = ((ab_totaltime - int_cycle_elapsed),time_complement)
        tuple5 += load3
    else:
        # If you are happy with tail being None,None,None on fail then no
        # need for special if / then enclosure - simple += append would do.
        tuple5 = ((ab_totaltime - int_cycle_elapsed),time_complement,0,0,0)
    assert len(tuple5) == 5
    return tuple5


def poor_cycle(target='/rc/py',concurrent=2,totaltime=10,
               data='initial',visible='internal',
               app_name='local2dj14',dir_fab='poor_cycle.fab',port_wsgi=None):
    """ See two functions setup and load for full function signature similar
    to target='/rc/py',concurrent=2,totaltime=180,data='initial',
               visible='internal',app_name='local2dj14',dir_fab='poor_cycle.fab'

    Poor config to deliberately stress a system for the purposes of
    benchmarking. Having 2 or more concurrent users requesting pygments
    prettified output (non cached) is usually enough to run a test benchmark
    against a low budget cloud server.
    A properly configured Apache mod_wsgi can work well, but in order to
    stress our system here, and to cause least disruption to any existing setup,
    we employ the minimum of Apache configuration effort
    An Apache configuration which has not been tailored to local conditions on
    the server is probably a poor configuration.
    Such configurations can introduce system stress and make a taxing benchmark.

    django14cycle differs from django14snail in that the concept of a basket or cycle of
    tasks is introduced. This is intended to introduce a single thread / non Apache measure
    into the final mark.
    Apache bench is still used, but as a contribution to, rather than the entire benchmark.

    Returns result from .._load() function which is a 5 tuple:
        tuple5 = ((ab_totaltime - int_cycle_elapsed),time_complement,0,0,0)
    Last three zeros would normally contain load averages rather than zeros 
    """
    tuple5 = none5 = (None,None,None,None,None)
    swap_total = cmdbin.free_marked(True)
    """ Having reported small amount of summary information about the server,
    next we have two steps to complete: 'setup' then 'load'
    """
    dirpath_django_project,dirpath_django_log,dirpath_djapian = poor_cycle_setup14(
        target,concurrent,totaltime,data,visible,app_name,dir_fab,port_wsgi)

    if dirpath_django_project is not None:
        tuple5 = poor_cycle_load14(target,concurrent,totaltime,dir_fab,
                                   dirpath_django_project,dirpath_django_log,
                                   dirpath_djapian,port_wsgi)
        if any(tuple5):
            pass
        else:
            tuple5 = none5
    return tuple5


def sync_load_index14one(print_flag=False,port_wsgi=None):
    """ Perform a single sync, load, index and optionally print the result.

    How destructive / intrusive is running this routine?
    We certainly install apache2 if it is not already installed.
    We do not change the existing MPM - it stays as it is when apache2 already present
    We do change apache2 to listen additionally on port 1480

    Could you run this routine on a live server? Probably.
    Worried about an existing apache2 setup? Take a recursive cp of /etc/apache2
    before running and revert the opening of 1480 port manually once output has completed.

    Rather than use poor_cycle_setup14, we could in future create a variant that
    does not add apache2 ports whilst doing the Django setup. However for now we are
    reusing existing code that is already in place.
    """
    dirpath_django_project,dirpath_django_log,dirpath_djapian = poor_cycle_setup14(
        '/rc/py',concurrent=1,totaltime=5,data='initial',visible='internal',
        app_name='local2dj14',dir_fab='sync_load_index.fab',port_wsgi=port_wsgi)

    seconds = 0
    if dirpath_django_project is not None:
        seconds = sync_load_index(dirpath_django_project,dirpath_django_log,dirpath_djapian,print_flag)
    return seconds

